@@ -17,11 +17,10 @@
 package com.android.apksig;
 
 import com.android.apksig.apk.ApkFormatException;
+import com.android.apksig.apk.ApkSigningBlockNotFoundException;
 import com.android.apksig.apk.ApkUtils;
 import com.android.apksig.apk.MinSdkVersionException;
-import com.android.apksig.internal.apk.v2.V2SchemeVerifier;
 import com.android.apksig.internal.util.ByteBufferDataSource;
-import com.android.apksig.internal.util.Pair;
 import com.android.apksig.internal.zip.CentralDirectoryRecord;
 import com.android.apksig.internal.zip.EocdRecord;
 import com.android.apksig.internal.zip.LocalFileRecord;
@@ -32,7 +31,9 @@ import com.android.apksig.util.DataSource;
 import com.android.apksig.util.DataSources;
 import com.android.apksig.util.ReadableDataSink;
 import com.android.apksig.zip.ZipFormatException;
+import java.io.ByteArrayOutputStream;
 import java.io.Closeable;
+import java.io.DataOutputStream;
 import java.io.File;
 import java.io.IOException;
 import java.io.RandomAccessFile;
@@ -50,6 +51,7 @@ import java.util.HashSet;
 import java.util.List;
 import java.util.Map;
 import java.util.Set;
+import java.util.regex.Pattern;
 
 /**
  * APK signer.
@@ -77,6 +79,8 @@ public class ApkSigner {
      */
     private static final short ALIGNMENT_ZIP_EXTRA_DATA_FIELD_MIN_SIZE_BYTES = 6;
 
+    private static final short ANDROID_COMMON_PAGE_ALIGNMENT_BYTES = 4096;
+
     /**
      * Name of the Android manifest ZIP entry in APKs.
      */
@@ -86,6 +90,8 @@ public class ApkSigner {
     private final Integer mMinSdkVersion;
     private final boolean mV1SigningEnabled;
     private final boolean mV2SigningEnabled;
+    private final boolean mV3SigningEnabled;
+    private final boolean mDebuggableApkPermitted;
     private final boolean mOtherSignersSignaturesPreserved;
     private final String mCreatedBy;
 
@@ -98,11 +104,15 @@ public class ApkSigner {
     private final DataSink mOutputApkDataSink;
     private final DataSource mOutputApkDataSource;
 
+    private final SigningCertificateLineage mSigningCertificateLineage;
+
     private ApkSigner(
             List<SignerConfig> signerConfigs,
             Integer minSdkVersion,
             boolean v1SigningEnabled,
             boolean v2SigningEnabled,
+            boolean v3SigningEnabled,
+            boolean debuggableApkPermitted,
             boolean otherSignersSignaturesPreserved,
             String createdBy,
             ApkSignerEngine signerEngine,
@@ -110,12 +120,15 @@ public class ApkSigner {
             DataSource inputApkDataSource,
             File outputApkFile,
             DataSink outputApkDataSink,
-            DataSource outputApkDataSource) {
+            DataSource outputApkDataSource,
+            SigningCertificateLineage signingCertificateLineage) {
 
         mSignerConfigs = signerConfigs;
         mMinSdkVersion = minSdkVersion;
         mV1SigningEnabled = v1SigningEnabled;
         mV2SigningEnabled = v2SigningEnabled;
+        mV3SigningEnabled = v3SigningEnabled;
+        mDebuggableApkPermitted = debuggableApkPermitted;
         mOtherSignersSignaturesPreserved = otherSignersSignaturesPreserved;
         mCreatedBy = createdBy;
 
@@ -127,6 +140,8 @@ public class ApkSigner {
         mOutputApkFile = outputApkFile;
         mOutputApkDataSink = outputApkDataSink;
         mOutputApkDataSource = outputApkDataSource;
+
+        mSigningCertificateLineage = signingCertificateLineage;
     }
 
     /**
@@ -204,14 +219,14 @@ public class ApkSigner {
         long inputApkSigningBlockOffset = -1;
         DataSource inputApkSigningBlock = null;
         try {
-            Pair<DataSource, Long> apkSigningBlockAndOffset =
-                    V2SchemeVerifier.findApkSigningBlock(inputApk, inputZipSections);
-            inputApkSigningBlock = apkSigningBlockAndOffset.getFirst();
-            inputApkSigningBlockOffset = apkSigningBlockAndOffset.getSecond();
-        } catch (V2SchemeVerifier.SignatureNotFoundException e) {
+            ApkUtils.ApkSigningBlock apkSigningBlockInfo =
+                    ApkUtils.findApkSigningBlock(inputApk, inputZipSections);
+            inputApkSigningBlockOffset = apkSigningBlockInfo.getStartOffset();
+            inputApkSigningBlock = apkSigningBlockInfo.getContents();
+        } catch (ApkSigningBlockNotFoundException e) {
             // Input APK does not contain an APK Signing Block. That's OK. APKs are not required to
             // contain this block. It's only needed if the APK is signed using APK Signature Scheme
-            // v2.
+            // v2 and/or v3.
         }
         DataSource inputApkLfhSection =
                 inputApk.slice(
@@ -225,6 +240,9 @@ public class ApkSigner {
         List<CentralDirectoryRecord> inputCdRecords =
                 parseZipCentralDirectory(inputCd, inputZipSections);
 
+        List<Pattern> pinPatterns = extractPinPatterns(inputCdRecords, inputApkLfhSection);
+        List<Hints.ByteRange> pinByteRanges = pinPatterns == null ? null : new ArrayList<>();
+
         // Step 3. Obtain a signer engine instance
         ApkSignerEngine signerEngine;
         if (mSignerEngine != null) {
@@ -254,7 +272,10 @@ public class ApkSigner {
                     new DefaultApkSignerEngine.Builder(engineSignerConfigs, minSdkVersion)
                             .setV1SigningEnabled(mV1SigningEnabled)
                             .setV2SigningEnabled(mV2SigningEnabled)
-                            .setOtherSignersSignaturesPreserved(mOtherSignersSignaturesPreserved);
+                            .setV3SigningEnabled(mV3SigningEnabled)
+                            .setDebuggableApkPermitted(mDebuggableApkPermitted)
+                            .setOtherSignersSignaturesPreserved(mOtherSignersSignaturesPreserved)
+                            .setSigningCertificateLineage(mSigningCertificateLineage);
             if (mCreatedBy != null) {
                 signerEngineBuilder.setCreatedBy(mCreatedBy);
             }
@@ -283,6 +304,9 @@ public class ApkSigner {
                 new HashMap<>(inputCdRecords.size());
         for (final CentralDirectoryRecord inputCdRecord : inputCdRecordsSortedByLfhOffset) {
             String entryName = inputCdRecord.getName();
+            if (Hints.PIN_BYTE_RANGE_ZIP_ENTRY_NAME.equals(entryName)) {
+                continue;  // We'll re-add below if needed.
+            }
             ApkSignerEngine.InputJarEntryInstructions entryInstructions =
                     signerEngine.inputJarEntry(entryName);
             boolean shouldOutput;
@@ -355,6 +379,23 @@ public class ApkSigner {
                                 outputLocalFileHeaderOffset);
                 outputOffset += outputLocalFileRecordSize;
 
+                if (pinPatterns != null) {
+                    boolean pinThisFile = false;
+                    for (Pattern pinPattern : pinPatterns) {
+                        if (pinPattern.matcher(inputCdRecord.getName()).matches()) {
+                            pinThisFile = true;
+                            break;
+                        }
+                    }
+
+                    if (pinThisFile) {
+                        pinByteRanges.add(
+                            new Hints.ByteRange(
+                                outputLocalFileHeaderOffset,
+                                outputOffset));
+                    }
+                }
+
                 // Enqueue entry's Central Directory record for output
                 CentralDirectoryRecord outputCdRecord;
                 if (outputLocalFileHeaderOffset == inputLocalFileRecord.getStartOffsetInArchive()) {
@@ -440,6 +481,35 @@ public class ApkSigner {
             outputJarSignatureRequest.done();
         }
 
+        if (pinByteRanges != null) {
+            pinByteRanges.add(new Hints.ByteRange(outputOffset, Long.MAX_VALUE));  // central dir
+            String entryName = Hints.PIN_BYTE_RANGE_ZIP_ENTRY_NAME;
+            byte[] uncompressedData = Hints.encodeByteRangeList(pinByteRanges);
+            ZipUtils.DeflateResult deflateResult =
+                    ZipUtils.deflate(ByteBuffer.wrap(uncompressedData));
+            byte[] compressedData = deflateResult.output;
+            long uncompressedDataCrc32 = deflateResult.inputCrc32;
+            long localFileHeaderOffset = outputOffset;
+            outputOffset +=
+                    LocalFileRecord.outputRecordWithDeflateCompressedData(
+                        entryName,
+                        lastModifiedTimeForNewEntries,
+                        lastModifiedDateForNewEntries,
+                        compressedData,
+                        uncompressedDataCrc32,
+                        uncompressedData.length,
+                        outputApkOut);
+            outputCdRecords.add(
+                CentralDirectoryRecord.createWithDeflateCompressedData(
+                    entryName,
+                    lastModifiedTimeForNewEntries,
+                    lastModifiedDateForNewEntries,
+                    uncompressedDataCrc32,
+                    compressedData.length,
+                    uncompressedData.length,
+                    localFileHeaderOffset));
+        }
+
         // Step 8. Construct output ZIP Central Directory in an in-memory buffer
         long outputCentralDirSizeBytes = 0;
         for (CentralDirectoryRecord record : outputCdRecords) {
@@ -467,19 +537,22 @@ public class ApkSigner {
                         outputCentralDirDataSource.size(),
                         outputCentralDirStartOffset);
 
-        // Step 10. Generate and output APK Signature Scheme v2 signatures, if necessary. This may
-        // insert an APK Signing Block just before the output's ZIP Central Directory
-        ApkSignerEngine.OutputApkSigningBlockRequest outputApkSigingBlockRequest =
-                signerEngine.outputZipSections(
+        // Step 10. Generate and output APK Signature Scheme v2 and/or v3 signatures, if necessary.
+        // This may insert an APK Signing Block just before the output's ZIP Central Directory
+        ApkSignerEngine.OutputApkSigningBlockRequest2 outputApkSigningBlockRequest =
+                signerEngine.outputZipSections2(
                         outputApkIn,
                         outputCentralDirDataSource,
                         DataSources.asDataSource(outputEocd));
-        if (outputApkSigingBlockRequest != null) {
-            byte[] outputApkSigningBlock = outputApkSigingBlockRequest.getApkSigningBlock();
+
+        if (outputApkSigningBlockRequest != null) {
+            int padding = outputApkSigningBlockRequest.getPaddingSizeBeforeApkSigningBlock();
+            outputApkOut.consume(ByteBuffer.allocate(padding));
+            byte[] outputApkSigningBlock = outputApkSigningBlockRequest.getApkSigningBlock();
             outputApkOut.consume(outputApkSigningBlock, 0, outputApkSigningBlock.length);
-            ZipUtils.setZipEocdCentralDirectoryOffset(
-                    outputEocd, outputCentralDirStartOffset + outputApkSigningBlock.length);
-            outputApkSigingBlockRequest.done();
+            ZipUtils.setZipEocdCentralDirectoryOffset(outputEocd,
+                    outputCentralDirStartOffset + padding + outputApkSigningBlock.length);
+            outputApkSigningBlockRequest.done();
         }
 
         // Step 11. Output ZIP Central Directory and ZIP End of Central Directory
@@ -577,7 +650,7 @@ public class ApkSigner {
         }
 
         // Fall back to filename-based defaults
-        return (entry.getName().endsWith(".so")) ? 4096 : 4;
+        return (entry.getName().endsWith(".so")) ? ANDROID_COMMON_PAGE_ALIGNMENT_BYTES : 4;
     }
 
     private static ByteBuffer createExtraFieldToAlignData(
@@ -689,6 +762,16 @@ public class ApkSigner {
         return cdRecords;
     }
 
+    private static CentralDirectoryRecord findCdRecord(
+        List<CentralDirectoryRecord> cdRecords, String name) {
+        for (CentralDirectoryRecord cdRecord : cdRecords) {
+            if (name.equals(cdRecord.getName())) {
+                return cdRecord;
+            }
+        }
+        return null;
+    }
+
     /**
      * Returns the contents of the APK's {@code AndroidManifest.xml} or {@code null} if this entry
      * is not present in the APK.
@@ -696,13 +779,8 @@ public class ApkSigner {
     static ByteBuffer getAndroidManifestFromApk(
             List<CentralDirectoryRecord> cdRecords, DataSource lhfSection)
                     throws IOException, ApkFormatException, ZipFormatException {
-        CentralDirectoryRecord androidManifestCdRecord = null;
-        for (CentralDirectoryRecord cdRecord : cdRecords) {
-            if (ANDROID_MANIFEST_ZIP_ENTRY_NAME.equals(cdRecord.getName())) {
-                androidManifestCdRecord = cdRecord;
-                break;
-            }
-        }
+        CentralDirectoryRecord androidManifestCdRecord =
+                findCdRecord(cdRecords, ANDROID_MANIFEST_ZIP_ENTRY_NAME);
         if (androidManifestCdRecord == null) {
             throw new ApkFormatException("Missing " + ANDROID_MANIFEST_ZIP_ENTRY_NAME);
         }
@@ -712,6 +790,30 @@ public class ApkSigner {
                         lhfSection, androidManifestCdRecord, lhfSection.size()));
     }
 
+    /**
+     * Return list of pin patterns embedded in the pin pattern asset
+     * file.  If no such file, return {@code null}.
+     */
+    private static List<Pattern> extractPinPatterns(
+            List<CentralDirectoryRecord> cdRecords, DataSource lhfSection)
+                    throws IOException, ApkFormatException {
+        CentralDirectoryRecord pinListCdRecord =
+                findCdRecord(cdRecords, Hints.PIN_HINT_ASSET_ZIP_ENTRY_NAME);
+        List<Pattern> pinPatterns = null;
+        if (pinListCdRecord != null) {
+            pinPatterns = new ArrayList<>();
+            byte[] patternBlob;
+            try {
+                patternBlob = LocalFileRecord.getUncompressedData(
+                    lhfSection, pinListCdRecord, lhfSection.size());
+            } catch (ZipFormatException ex) {
+                throw new ApkFormatException("Bad " + pinListCdRecord);
+            }
+            pinPatterns = Hints.parsePinPatterns(patternBlob);
+        }
+        return pinPatterns;
+    }
+
     /**
      * Returns the minimum Android version (API Level) supported by the provided APK. This is based
      * on the {@code android:minSdkVersion} attributes of the APK's {@code AndroidManifest.xml}.
@@ -820,7 +922,8 @@ public class ApkSigner {
      * <ul>
      * <li>Signer configs or {@link ApkSignerEngine} -- provided in the constructor,</li>
      * <li>APK to be signed -- see {@link #setInputApk(File) setInputApk} variants,</li>
-     * <li>where to store the signed APK -- see {@link #setOutputApk(File) setOutputApk} variants.
+     * <li>where to store the output signed APK -- see {@link #setOutputApk(File) setOutputApk}
+     *     variants.
      * </li>
      * </ul>
      */
@@ -828,6 +931,8 @@ public class ApkSigner {
         private final List<SignerConfig> mSignerConfigs;
         private boolean mV1SigningEnabled = true;
         private boolean mV2SigningEnabled = true;
+        private boolean mV3SigningEnabled = true;
+        private boolean mDebuggableApkPermitted = true;
         private boolean mOtherSignersSignaturesPreserved;
         private String mCreatedBy;
         private Integer mMinSdkVersion;
@@ -841,6 +946,17 @@ public class ApkSigner {
         private DataSink mOutputApkDataSink;
         private DataSource mOutputApkDataSource;
 
+        private SigningCertificateLineage mSigningCertificateLineage;
+
+        // APK Signature Scheme v3 only supports a single signing certificate, so to move to v3
+        // signing by default, but not require prior clients to update to explicitly disable v3
+        // signing for multiple signers, we modify the mV3SigningEnabled depending on the provided
+        // inputs (multiple signers and mSigningCertificateLineage in particular).  Maintain two
+        // extra variables to record whether or not mV3SigningEnabled has been set directly by a
+        // client and so should override the default behavior.
+        private boolean mV3SigningExplicitlyDisabled = false;
+        private boolean mV3SigningExplicitlyEnabled = false;
+
         /**
          * Constructs a new {@code Builder} for an {@code ApkSigner} which signs using the provided
          * signer configurations. The resulting signer may be further customized through this
@@ -855,6 +971,12 @@ public class ApkSigner {
             if (signerConfigs.isEmpty()) {
                 throw new IllegalArgumentException("At least one signer config must be provided");
             }
+            if (signerConfigs.size() > 1) {
+                // APK Signature Scheme v3 only supports single signer, unless a
+                // SigningCertificateLineage is provided, in which case this will be reset to true,
+                // since we don't yet have a v4 scheme about which to worry
+                mV3SigningEnabled = false;
+            }
             mSignerConfigs = new ArrayList<>(signerConfigs);
             mSignerEngine = null;
         }
@@ -1032,6 +1154,55 @@ public class ApkSigner {
             return this;
         }
 
+        /**
+         * Sets whether the APK should be signed using APK Signature Scheme v3 (aka v3 signature
+         * scheme).
+         *
+         * <p>By default, whether APK is signed using APK Signature Scheme v3 is determined by
+         * {@code ApkSigner} based on the platform versions supported by the APK or specified using
+         * {@link #setMinSdkVersion(int)}.
+         *
+         * <p><em>Note:</em> This method may only be invoked when this builder is not initialized
+         * with an {@link ApkSignerEngine}.
+         * <p><em>Note:</em> APK Signature Scheme v3 only supports a single signing certificate, but
+         * may take multiple signers mapping to different targeted platform versions.
+         *
+         * @param enabled {@code true} to require the APK to be signed using APK Signature Scheme
+         *        v3, {@code false} to require the APK to not be signed using APK Signature Scheme
+         *        v3.
+         *
+         * @throws IllegalStateException if this builder was initialized with an
+         *         {@link ApkSignerEngine}
+         */
+        public Builder setV3SigningEnabled(boolean enabled) {
+            checkInitializedWithoutEngine();
+            mV3SigningEnabled = enabled;
+            if (enabled) {
+                mV3SigningExplicitlyEnabled = true;
+            } else {
+                mV3SigningExplicitlyDisabled = true;
+            }
+            return this;
+        }
+
+        /**
+         * Sets whether the APK should be signed even if it is marked as debuggable
+         * ({@code android:debuggable="true"} in its {@code AndroidManifest.xml}). For backward
+         * compatibility reasons, the default value of this setting is {@code true}.
+         *
+         * <p>It is dangerous to sign debuggable APKs with production/release keys because Android
+         * platform loosens security checks for such APKs. For example, arbitrary unauthorized code
+         * may be executed in the context of such an app by anybody with ADB shell access.
+         *
+         * <p><em>Note:</em> This method may only be invoked when this builder is not initialized
+         * with an {@link ApkSignerEngine}.
+         */
+        public Builder setDebuggableApkPermitted(boolean permitted) {
+            checkInitializedWithoutEngine();
+            mDebuggableApkPermitted = permitted;
+            return this;
+        }
+
         /**
          * Sets whether signatures produced by signers other than the ones configured in this engine
          * should be copied from the input APK to the output APK.
@@ -1075,16 +1246,48 @@ public class ApkSigner {
             }
         }
 
+        /**
+         * Sets the {@link SigningCertificateLineage} to use with the v3 signature scheme.  This
+         * structure provides proof of signing certificate rotation linking {@link SignerConfig}
+         * objects to previous ones.
+         */
+        public Builder setSigningCertificateLineage(
+                SigningCertificateLineage signingCertificateLineage) {
+            if (signingCertificateLineage != null) {
+                mV3SigningEnabled = true;
+                mSigningCertificateLineage = signingCertificateLineage;
+            }
+            return this;
+        }
+
         /**
          * Returns a new {@code ApkSigner} instance initialized according to the configuration of
          * this builder.
          */
         public ApkSigner build() {
+
+            if (mV3SigningExplicitlyDisabled && mV3SigningExplicitlyEnabled) {
+                throw new IllegalStateException("Builder configured to both enable and disable APK "
+                        + "Signature Scheme v3 signing");
+            }
+
+            if (mV3SigningExplicitlyDisabled) {
+                mV3SigningEnabled = false;
+            }
+
+            if (mV3SigningExplicitlyEnabled) {
+                mV3SigningEnabled = true;
+            }
+
+            // TODO - if v3 signing is enabled, check provided signers and history to see if valid
+
             return new ApkSigner(
                     mSignerConfigs,
                     mMinSdkVersion,
                     mV1SigningEnabled,
                     mV2SigningEnabled,
+                    mV3SigningEnabled,
+                    mDebuggableApkPermitted,
                     mOtherSignersSignaturesPreserved,
                     mCreatedBy,
                     mSignerEngine,
@@ -1092,7 +1295,8 @@ public class ApkSigner {
                     mInputApkDataSource,
                     mOutputApkFile,
                     mOutputApkDataSink,
-                    mOutputApkDataSource);
+                    mOutputApkDataSource,
+                    mSigningCertificateLineage);
         }
     }
 }
