@@ -37,14 +37,16 @@ import com.android.apksig.internal.pkcs7.SignerIdentifier;
 import com.android.apksig.internal.pkcs7.SignerInfo;
 import com.android.apksig.internal.util.AndroidSdkVersion;
 import com.android.apksig.internal.util.ByteBufferUtils;
+import com.android.apksig.internal.util.X509CertificateUtils;
 import com.android.apksig.internal.util.GuaranteedEncodedFormX509Certificate;
 import com.android.apksig.internal.util.InclusiveIntRange;
-import com.android.apksig.internal.util.MessageDigestSink;
+import com.android.apksig.internal.util.Pair;
 import com.android.apksig.internal.zip.CentralDirectoryRecord;
 import com.android.apksig.internal.zip.LocalFileRecord;
+import com.android.apksig.util.DataSinks;
 import com.android.apksig.util.DataSource;
 import com.android.apksig.zip.ZipFormatException;
-import java.io.ByteArrayInputStream;
+
 import java.io.IOException;
 import java.math.BigInteger;
 import java.nio.ByteBuffer;
@@ -56,7 +58,6 @@ import java.security.Principal;
 import java.security.Signature;
 import java.security.SignatureException;
 import java.security.cert.CertificateException;
-import java.security.cert.CertificateFactory;
 import java.security.cert.X509Certificate;
 import java.util.ArrayList;
 import java.util.Arrays;
@@ -72,6 +73,7 @@ import java.util.Map;
 import java.util.Set;
 import java.util.StringTokenizer;
 import java.util.jar.Attributes;
+
 import javax.security.auth.x500.X500Principal;
 
 /**
@@ -90,6 +92,12 @@ public abstract class V1SchemeVerifier {
      * considered verified only if {@link Result#verified} is {@code true}. If verification fails,
      * the result will contain errors -- see {@link Result#getErrors()}.
      *
+     * <p>Verification succeeds iff the APK's JAR signatures are expected to verify on all Android
+     * platform versions in the {@code [minSdkVersion, maxSdkVersion]} range. If the APK's signature
+     * is expected to not verify on any of the specified platform versions, this method returns a
+     * result with one or more errors and whose {@code Result.verified == false}, or this method
+     * throws an exception.
+     *
      * @throws ApkFormatException if the APK is malformed
      * @throws IOException if an I/O error occurs when reading the APK
      * @throws NoSuchAlgorithmException if the APK's JAR signatures cannot be verified because a
@@ -155,6 +163,44 @@ public abstract class V1SchemeVerifier {
         return cdEntryNames;
     }
 
+    /**
+    * Parses raw representation of MANIFEST.MF file into a pair of main entry manifest section
+    * representation and a mapping between entry name and its manifest section representation.
+    *
+    * @param manifestBytes raw representation of Manifest.MF
+    * @param cdEntryNames expected set of entry names
+    * @param result object to keep track of errors that happened during the parsing
+    * @return a pair of main entry manifest section representation and a mapping between entry name
+    *     and its manifest section representation
+    */
+    public static Pair<ManifestParser.Section, Map<String, ManifestParser.Section>> parseManifest(
+            byte[] manifestBytes, Set<String> cdEntryNames, Result result) {
+        ManifestParser manifest = new ManifestParser(manifestBytes);
+        ManifestParser.Section manifestMainSection = manifest.readSection();
+        List<ManifestParser.Section> manifestIndividualSections = manifest.readAllSections();
+        Map<String, ManifestParser.Section> entryNameToManifestSection =
+                new HashMap<>(manifestIndividualSections.size());
+        int manifestSectionNumber = 0;
+        for (ManifestParser.Section manifestSection : manifestIndividualSections) {
+            manifestSectionNumber++;
+            String entryName = manifestSection.getName();
+            if (entryName == null) {
+                result.addError(Issue.JAR_SIG_UNNNAMED_MANIFEST_SECTION, manifestSectionNumber);
+                continue;
+            }
+            if (entryNameToManifestSection.put(entryName, manifestSection) != null) {
+                result.addError(Issue.JAR_SIG_DUPLICATE_MANIFEST_SECTION, entryName);
+                continue;
+            }
+            if (!cdEntryNames.contains(entryName)) {
+                result.addError(
+                        Issue.JAR_SIG_MISSING_ZIP_ENTRY_REFERENCED_IN_MANIFEST, entryName);
+                continue;
+            }
+        }
+        return Pair.of(manifestMainSection, entryNameToManifestSection);
+    }
+
     /**
      * All JAR signers of an APK.
      */
@@ -213,32 +259,18 @@ public abstract class V1SchemeVerifier {
             } catch (ZipFormatException e) {
                 throw new ApkFormatException("Malformed ZIP entry: " + manifestEntry.getName(), e);
             }
-            Map<String, ManifestParser.Section> entryNameToManifestSection = null;
-            ManifestParser manifest = new ManifestParser(manifestBytes);
-            ManifestParser.Section manifestMainSection = manifest.readSection();
-            List<ManifestParser.Section> manifestIndividualSections = manifest.readAllSections();
-            entryNameToManifestSection = new HashMap<>(manifestIndividualSections.size());
-            int manifestSectionNumber = 0;
-            for (ManifestParser.Section manifestSection : manifestIndividualSections) {
-                manifestSectionNumber++;
-                String entryName = manifestSection.getName();
-                if (entryName == null) {
-                    result.addError(Issue.JAR_SIG_UNNNAMED_MANIFEST_SECTION, manifestSectionNumber);
-                    continue;
-                }
-                if (entryNameToManifestSection.put(entryName, manifestSection) != null) {
-                    result.addError(Issue.JAR_SIG_DUPLICATE_MANIFEST_SECTION, entryName);
-                    continue;
-                }
-                if (!cdEntryNames.contains(entryName)) {
-                    result.addError(
-                            Issue.JAR_SIG_MISSING_ZIP_ENTRY_REFERENCED_IN_MANIFEST, entryName);
-                    continue;
-                }
-            }
+
+            Pair<ManifestParser.Section, Map<String, ManifestParser.Section>> manifestSections =
+                    parseManifest(manifestBytes, cdEntryNames, result);
+
             if (result.containsErrors()) {
                 return;
             }
+
+            ManifestParser.Section manifestMainSection = manifestSections.getFirst();
+            Map<String, ManifestParser.Section> entryNameToManifestSection =
+                    manifestSections.getSecond();
+
             // STATE OF AFFAIRS:
             // * All JAR entries listed in JAR manifest are present in the APK.
 
@@ -742,22 +774,13 @@ public abstract class V1SchemeVerifier {
                 return Collections.emptyList();
             }
 
-            CertificateFactory certFactory;
-            try {
-                certFactory = CertificateFactory.getInstance("X.509");
-            } catch (CertificateException e) {
-                throw new RuntimeException("Failed to create X.509 CertificateFactory", e);
-            }
-
             List<X509Certificate> result = new ArrayList<>(encodedCertificates.size());
             for (int i = 0; i < encodedCertificates.size(); i++) {
                 Asn1OpaqueObject encodedCertificate = encodedCertificates.get(i);
                 X509Certificate certificate;
                 byte[] encodedForm = ByteBufferUtils.toByteArray(encodedCertificate.getEncoded());
                 try {
-                    certificate =
-                            (X509Certificate) certFactory.generateCertificate(
-                                    new ByteArrayInputStream(encodedForm));
+                    certificate = X509CertificateUtils.generateCertificate(encodedForm);
                 } catch (CertificateException e) {
                     throw new CertificateException("Failed to parse certificate #" + (i + 1), e);
                 }
@@ -842,6 +865,8 @@ public abstract class V1SchemeVerifier {
         private static final String OID_SIG_SHA1_WITH_DSA = "1.2.840.10040.4.3";
         private static final String OID_SIG_SHA224_WITH_DSA = "2.16.840.1.101.3.4.3.1";
         static final String OID_SIG_SHA256_WITH_DSA = "2.16.840.1.101.3.4.3.2";
+        static final String OID_SIG_SHA384_WITH_DSA = "2.16.840.1.101.3.4.3.3";
+        static final String OID_SIG_SHA512_WITH_DSA = "2.16.840.1.101.3.4.3.4";
 
         static final String OID_SIG_EC_PUBLIC_KEY = "1.2.840.10045.2.1";
         private static final String OID_SIG_SHA1_WITH_ECDSA = "1.2.840.10045.4.1";
@@ -1209,6 +1234,8 @@ public abstract class V1SchemeVerifier {
                 OID_TO_USER_FRIENDLY_NAME.put(OID_SIG_SHA1_WITH_DSA, "SHA-1 with DSA");
                 OID_TO_USER_FRIENDLY_NAME.put(OID_SIG_SHA224_WITH_DSA, "SHA-224 with DSA");
                 OID_TO_USER_FRIENDLY_NAME.put(OID_SIG_SHA256_WITH_DSA, "SHA-256 with DSA");
+                OID_TO_USER_FRIENDLY_NAME.put(OID_SIG_SHA384_WITH_DSA, "SHA-384 with DSA");
+                OID_TO_USER_FRIENDLY_NAME.put(OID_SIG_SHA512_WITH_DSA, "SHA-512 with DSA");
 
                 OID_TO_USER_FRIENDLY_NAME.put(OID_SIG_EC_PUBLIC_KEY, "ECDSA");
                 OID_TO_USER_FRIENDLY_NAME.put(OID_SIG_SHA1_WITH_ECDSA, "SHA-1 with ECDSA");
@@ -1603,7 +1630,7 @@ public abstract class V1SchemeVerifier {
         }
     }
 
-    private static Collection<NamedDigest> getDigestsToVerify(
+    public static Collection<NamedDigest> getDigestsToVerify(
             ManifestParser.Section section,
             String digestAttrSuffix,
             int minSdkVersion,
@@ -1859,7 +1886,7 @@ public abstract class V1SchemeVerifier {
                         apk,
                         cdRecord,
                         cdOffsetInApk,
-                        new MessageDigestSink(mds));
+                        DataSinks.asDataSink(mds));
             } catch (ZipFormatException e) {
                 throw new ApkFormatException("Malformed ZIP entry: " + entryName, e);
             } catch (IOException e) {
@@ -1916,9 +1943,9 @@ public abstract class V1SchemeVerifier {
         return getMessageDigest(algorithm).digest(data);
     }
 
-    private static class NamedDigest {
-        private final String jcaDigestAlgorithm;
-        private final byte[] digest;
+    public static class NamedDigest {
+        public final String jcaDigestAlgorithm;
+        public final byte[] digest;
 
         private NamedDigest(String jcaDigestAlgorithm, byte[] digest) {
             this.jcaDigestAlgorithm = jcaDigestAlgorithm;
