@@ -22,6 +22,9 @@ import org.junit.runners.Suite;
 @RunWith(Suite.class)
 @Suite.SuiteClasses({
     ArrayBackedByteBufferSinkTest.class,
+    ChainedDataSourceTest.class,
     DirectByteBufferSinkTest.class,
+    VerityTreeBuilderTest.class,
+    X509CertificateUtilsTest.class,
 })
 public class AllTests {}
